/*
 * Transition-RRT.cpp
 *
 *  Created on: Nov 20, 2012
 *      Author: ddevaurs
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/Diffusion/Variants/Transition-RRT.hpp"

using namespace std;


/**
 * ****************************************************
 *      Transition Expansion
 * ****************************************************
 */

/**
 * Constructor.
 */
TransitionExpansion::TransitionExpansion(Graph * graph, Node * goal) : RRTExpansion(graph, goal)
{
	// set the temperature
	temperature = ENV.getDouble(Env::initialTemperature);

	// set the lowest and highest costs
	lowestCost = numeric_limits<double>::max();
	highestCost = -1. * numeric_limits<double>::max();
	vector<Node *> nodes = graph->getNodes();
	for (vector<Node *>::iterator n = nodes.begin(); n != nodes.end(); ++n)
		this->updateCostRange((*n)->cost());
}


/**
 * Attempt to link two given nodes (i.e. two components, when relevant).
 * @return TRUE if the junction is successful.
 */
bool TransitionExpansion::attemptJunction(Node * fromNode, Node * toNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
    confPtr_t toConf = toNode->getConfiguration();

    // path along which the junction is attempted
    LocalPath path(fromConf, toConf);
    double pathLength = path.getParamMax();

    // N.B.: the connection is attempted only if the path length is less than a specified threshold
    // (which is some multiple of the extension step)
    if (pathLength <= ENV.getDouble(Env::minimalFinalExpansionGap) * this->step())
    {
        // check whether the path is valid (i.e. collision-free)
        // but only if collisions have to be checked before the cost condition
        if (!ENV.getBool(Env::costBeforeColl))
            if (!path.isValid())
                return false;

        // interpolation factor and intermediate configurations
        double k = (pathLength == 0.) ? 1. : min(1., this->step() / pathLength);
        double prop(k);
        confPtr_t currentConf = fromConf;
        confPtr_t newConf = (prop == 1.) ? toConf : path.configAtParam(prop * pathLength);

        // cost condition: does the path follow a downhill slope in the cost space?
        while ((prop <= 1.) && (newConf->cost() <= currentConf->cost()))
        {
            if (prop < 1.)
            {
                prop = min(1., prop + k);
                currentConf = newConf;
                newConf = (prop == 1.) ? toConf : path.configAtParam(prop * pathLength);
            }
            else
                prop = 2.;
        }

        // if the cost condition is satisfied
        if (prop > 1.)
        {
            // check whether the path is valid (i.e. collision-free)
            // but only if collisions have to be checked after the cost condition
            if (ENV.getBool(Env::costBeforeColl))
                if (!path.isValid())
                    return false;

            // validate the junction
			return true;
        }
    }

    return false;
}



/**
 * T-RRT Extend.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned TransitionExpansion::extend(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
    ConnectedComponent * fromComp = fromNode->getConnectedComponent();

    // path along which the extension is performed
    LocalPath directionPath(fromConf, directionConf);
    double k = this->computeInterpolationFactor(directionPath.getParamMax());

    // Refinement control:
    // (k == 1. && !directionNode) means that we are potentially about to add a new refinement node;
    // in that case, if the number of refinement nodes already present in the component is too high
    // (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
    if (ENV.getBool(Env::refinementControl) && (k == 1. && !directionNode) &&
            (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
        return 0;

    // construct an 'extension path', by linear interpolation along the 'direction path'
    LocalPath extensionPath(fromConf, this->interpolate(directionConf, &directionPath, k));
    confPtr_t newConf = extensionPath.getEnd();

    // check cost, and then collisions
    if (ENV.getBool(Env::costBeforeColl))
    {
        // perform the transition test
        if (!this->transitionTest(fromConf, newConf))
            return 0;

        // check whether the extension path is valid (i.e. without collision)
        if (!extensionPath.isValid())
            return 0;
    }
    // OR check collisions, and then cost
    else
    {
        // check whether the extension path is valid (i.e. without collision)
        if (!extensionPath.isValid())
            return 0;

        // perform the transition test
        if (!this->transitionTest(fromConf, newConf))
            return 0;
    }

    // if everything succeeds, add a new node to the graph
    unsigned nbCreatedNodes(0);
    this->addNode(fromNode, extensionPath, k, directionNode, nbCreatedNodes);

    if (nbCreatedNodes > 0)
    {
		// update the cost range, based on the cost of the new node
		this->updateCostRange(newConf->cost());

		// Refinement control:
		// if a new refinement node has been built, update the number of refinement nodes of the component
		if (ENV.getBool(Env::refinementControl) && (k == 1.))
			fromComp->updateNumberOfRefinementNodes(nbCreatedNodes);
    }

	return nbCreatedNodes;
}


/**
 * T-RRT Connect.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned TransitionExpansion::connect(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
    ConnectedComponent * fromComp = fromNode->getConnectedComponent();

    // path along which the connection is performed
    LocalPath directionPath(fromConf, directionConf);
    double k = this->computeInterpolationFactor(directionPath.getParamMax());

    // Refinement control:
    // (k == 1. && !directionNode) means that we are potentially about to add a new refinement node;
    // in that case, if the number of refinement nodes already present in the component is too high
    // (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
    if (ENV.getBool(Env::refinementControl) && (k == 1. && !directionNode) &&
            (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
        return 0;

    bool success(true);
    confPtr_t currentConf = fromConf;
    confPtr_t newConf;
    double prop(k);
    double oldProp(0.);

    // attempt the connection step by step
    while (success && prop <= 1.)
    {
        // check whether the current extension path is valid (i.e. without collision)
        newConf = this->interpolate(directionConf, &directionPath, prop);
        LocalPath extensionPath(currentConf, newConf);
        if (!extensionPath.isValid())
            success = false;

        // perform the transition test
        if (success && this->transitionTest(currentConf, newConf))
        {
            currentConf = newConf;
            oldProp = prop;
            if (prop < 1.)
                prop = min(1., prop + k);
            else
                prop = 2.;
        }
        else
            success = false;
    }

    // if "reached" or "advanced", add a node to the graph
    if (success || prop > k)
    {
    	unsigned nbCreatedNodes(0);
        LocalPath connectPath(fromConf, currentConf);
        this->addNode(fromNode, connectPath, oldProp, directionNode, nbCreatedNodes);

        if (nbCreatedNodes > 0)
        {
			// update the cost range, based on the cost of the new node
			this->updateCostRange(currentConf->cost());

			// Refinement control:
			// if a refinement node has been built, update the number of refinement nodes of the component
			if (ENV.getBool(Env::refinementControl) && (k == 1.))
				fromComp->updateNumberOfRefinementNodes(nbCreatedNodes);
        }

        return nbCreatedNodes;
    }
    else
    	return 0;
}

/**
 * Transition test between fromConf and toConf, with temperature tuning.
 * @return TRUE if the transition is accepted
 */
bool TransitionExpansion::transitionTest(confPtr_t fromConf, confPtr_t toConf)
{
	double fromCost = fromConf->cost();
	double toCost = toConf->cost();

    // failure, when cost is higher than maximal value
    if ((ENV.getDouble(Env::costMax) > 0.) && (toCost > ENV.getDouble(Env::costMax)))
        return false;

    // success, for down-hill or flat slopes
    if (toCost <= fromCost)
        return true;

    // Metropolis test
    if (exp((fromCost - toCost) / temperature) > 0.5)
    {
        // if success, decrease the temperature
        double costRange = max(1., highestCost - lowestCost);
        temperature /= pow(2, (toCost - fromCost) / (0.1 * costRange));
        if (temperature < ENV.getDouble(Env::initialTemperature))
        	temperature = ENV.getDouble(Env::initialTemperature);
        return true;
    }
    else
    {
        // if failure, increase the temperature
        temperature *= pow(2, ENV.getDouble(Env::temperatureRate));
        return false;
    }
}


/**
 * ****************************************************
 *      Transition RRT
 * ****************************************************
 */

/**
 * Initialize the T-RRT.
 * @return the number of nodes added to the graph
 */
unsigned TransitionRRT::init()
{
	// create the T-RRT expansion
    expansion = new TransitionExpansion(_Graph, _Goal);

    // initialize the Tree Planner
    return TreePlanner::init();
}
